﻿///////////////////////////////////////
//
//  Copyright 2014-2016 Brett Faulds
//
///////////////////////////////////////

using System;
using System.Collections.Generic;
using System.IO;
using UnityEditor;
using UnityEngine;
using System.Collections;
using Boomlagoon.JSON;

public class TetrisEditor : EditorWindow
{
	private enum TetrisMenu
	{
		ShapeEditor = 0,
		Grid
	};

	private TetrisMenu currentMenu;

	private const string SHAPES_DIR = "/Resources/Shapes";
	private const string SHAPE_EXT = ".txt";
	private string _dataPath;

	private Shape _viewingShape;
	private List<Shape> _loadedShapes = new List<Shape>();
	private Vector2 scrollView = new Vector2();

	private string _shapeInputField = string.Empty;
	private string _shapeNameInputField = string.Empty;
	private string _lastStatus = string.Empty;

	private float _timeout = 0;

	[MenuItem("Tetris/Editor %G")]
	public static void Init()
	{
		TetrisEditor window = GetWindow<TetrisEditor>();
		window.Show();
#if UNITY_4_6 || UNITY_4_7
		window.title = "Tetris Editor";
#else
        window.titleContent = new GUIContent("Tetris Editor");
#endif
		window.Initialize();
	}

	public void Initialize()
	{
		_dataPath = Application.dataPath + SHAPES_DIR;
		string[] files = Directory.GetFiles(_dataPath, "*.txt");
		foreach (string s in files)
		{
			using (StreamReader sr = new StreamReader(s))
			{
				string name = sr.ReadLine();
				string data = sr.ReadToEnd();
				_loadedShapes.Add(new Shape(JSONArray.Parse(data), name));
			}
		}
	}

	private void OnGUI()
	{
		GUILayout.Label("Current Menu: " + currentMenu, EditorStyles.boldLabel);

		GUILayout.BeginHorizontal();
		if (GUILayout.Button("Shape Editor"))
		{
			currentMenu = TetrisMenu.ShapeEditor;
		}
		GUILayout.EndHorizontal();

		switch (currentMenu)
		{
			case TetrisMenu.ShapeEditor:
				GUILayout.BeginArea(new Rect(50, 175, 150, position.height - 200));
				if (_viewingShape != null)
				{
					for (int i = 0; i < _viewingShape.Size; ++i)
					{
						GUILayout.BeginHorizontal();
						for (int x = 0; x < _viewingShape.Size; ++x)
						{
							GUILayout.Box("[" + _viewingShape.ShapeLines[i, x].ToString() + "]", EditorStyles.whiteLabel);
						}
						GUILayout.EndHorizontal();
					}
				}
				GUILayout.EndArea();

				GUILayout.Label("Shape Name", EditorStyles.boldLabel);
				_shapeNameInputField = GUILayout.TextField(_shapeNameInputField);
				GUILayout.Label("Shape JSON (i.e. [[0,1,0], [1,1,1]])", EditorStyles.boldLabel);
				GUILayout.BeginHorizontal();
				_shapeInputField = GUILayout.TextField(_shapeInputField);
				if (GUILayout.Button("Paste", GUILayout.Width(50))) //Paste any copied text to this text field
				{
					_shapeInputField = EditorGUIUtility.systemCopyBuffer;
				}
				GUILayout.EndHorizontal();

				if (File.Exists(_dataPath + "/" + _shapeNameInputField + SHAPE_EXT))
				{
					GUILayout.Label("Shape name already exists!", EditorStyles.boldLabel);
				}
				else if (_shapeNameInputField.Length > 0 && _shapeInputField.Length > 0 && GUILayout.Button("Create Shape", GUILayout.Width(position.width - 65)))
				{
					if (!string.IsNullOrEmpty(_shapeInputField))
					{
						JSONArray temp = JSONArray.Parse(_shapeInputField);
						if (temp != null)
						{
							if (temp.Length > Shape.MAX_SIZE)
							{
								_lastStatus = "Shape cannot exceed the MAX SIZE of " + Shape.MAX_SIZE;
								_timeout = Time.realtimeSinceStartup + 3;
								return;
							}

							foreach (JSONValue jv in temp)
							{
								if (jv.Array.Length > Shape.MAX_SIZE)
								{
									_lastStatus = "Shape cannot exceed the MAX SIZE of " + Shape.MAX_SIZE;
									_timeout = Time.realtimeSinceStartup + 3;
									return;
								}
							}

							string jsonData = temp.ToString();

							using (StreamWriter sw = new StreamWriter(_dataPath + "/" + _shapeNameInputField + SHAPE_EXT))
							{
								sw.WriteLine(_shapeNameInputField);
								sw.Write(jsonData);
							}

							_loadedShapes.Add(new Shape(temp, _shapeNameInputField));
							_viewingShape = _loadedShapes[_loadedShapes.Count - 1];
							_shapeInputField = string.Empty;
							_shapeNameInputField = string.Empty;
							_lastStatus = "Successful shape created";
							_timeout = Time.realtimeSinceStartup + 3;
							AssetDatabase.Refresh();
						}
					}
				}

				GUILayout.BeginArea(new Rect(250, 175, position.width - 250, position.height - 275));
				scrollView = GUILayout.BeginScrollView(scrollView);
				for (int i = _loadedShapes.Count - 1; i >= 0; --i)
				{
					GUILayout.BeginHorizontal();
					if (GUILayout.Button(_loadedShapes[i].Name))
					{
						_viewingShape = _loadedShapes[i];
					}
					if (GUILayout.Button("Delete Shape", GUILayout.Width(100)))
					{
						if (_viewingShape == _loadedShapes[i])
							_viewingShape = null;

						File.Delete(_dataPath + "/" + _loadedShapes[i].Name + SHAPE_EXT);
						_loadedShapes.Remove(_loadedShapes[i]);
						AssetDatabase.Refresh();
					}
					GUILayout.EndHorizontal();
				}
				GUILayout.EndScrollView();
				GUILayout.EndArea();
				
				//Simple status label at the bottom of the editor window
				GUILayout.BeginArea(new Rect(0, position.height - 15, position.width, 15));
				GUILayout.Label(_lastStatus);
				GUILayout.EndArea();
				
				break;
			case TetrisMenu.Grid:
				break;
		}
	}

	private void Update()
	{
		//Remove any status messages after timeout is finished
		if (_lastStatus.Length > 0)
		{
			if (_timeout < Time.realtimeSinceStartup)
			{
				_lastStatus = string.Empty;
			}

			Repaint();
		}
	}
}

///////////////////////////////////////
//
//  Copyright 2014-2016 Brett Faulds
//
///////////////////////////////////////