﻿////////////////////////////////////////////////////
//
//  Copyright 2014-2016 Brett Faulds
//
//  Project Example: Tetris
//  Purpose: To showcase how I would
//      program a tetris game.
//
//  Made in January 2015 for Unity 4.6
//  Total time spent: 10 Hours
//
//  This project features the following:
//  -A fully working Tetris Game
//  -Object Oriented Programming
//  -Basic Custom Editor Window
//  -Json being used to parse the blocks
//  -Multi-dimensional arrays
//  -Dynamic and code generated block images
//  -Web downloading
//  -Efficient coding an best practices
//
///////////////////////////////////////////////////

using System;
using System.Linq;
using Boomlagoon.JSON;
using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using UnityEngine.UI;

/// <summary>
/// The purpose of this class is to control all game/ui input.
/// 
/// Could be refined more to be put in seperate classes, 
/// but wanted to get it done in a day.
/// </summary>
public class TetrisController : MonoBehaviour
{
	public Text Score;
	public Text Highscores;
	public GameObject HighscoresGO;
	public GameObject GameOverGO;
	public Image NextPiece;
	public Image Logo;
	public ButtonHandler Rotate;
	public ButtonHandler MoveLeft;
	public ButtonHandler MoveRight;
	public ButtonHandler MoveDown;

	public float DelayBeforeMoveDown = 1;
	public float DelaySpeed = 1;
	public int OffsetX = 2;

	public List<TetrisLine> GameTetrisLines = new List<TetrisLine>();

	public AudioClip GameStart;
	public AudioClip GameOver;
	public AudioClip ClearedLine;

	private bool _isMovingLeft = false;
	private bool _isMovingRight = false;
	private bool _gameOver = false;

	private float _delayMoveDown = 1;
	private float _moveDelayTimer = 0.45f;

    public string LogoUrl = "https://db.tt/bALxdaJt"; //Image url to load as a logo
    public string HighscoresURL = "https://db.tt/ICsdRwP7"; //Highscores json file to load (i.e. { "HighScores: [ { "name": "Brett", "score": 10 }, { "name": "Brett F", "score": 20 } })

	private Shape[] _gameShapes; //All the shapes of the game will be loaded here
	private Shape _currentShape;
	private Shape _nextShape;
	private int _shapeOffsetX;
	private int _shapeOffsetY;
	private int _prevShapeID;
	private Color _prevShapeColor;

	private int _score = 0;
	private int _shapeID = 0;
	private const int DEFAULT_SHAPE_ID = -1;

	private List<ShapeObject> shapeParts = new List<ShapeObject>();
	private AudioSource _audioSource;

	private void Awake()
	{
		//Clear all the lines!
		foreach (TetrisLine line in GameTetrisLines)
		{
			line.ClearLine();
		}

		//Holding of all button events for mobile
		Rotate.ButtonHeld += RotateShape;
		MoveLeft.ButtonHeld += MoveLeftHeld;
		MoveLeft.ButtonReleased += MoveLeftReleased;
		MoveRight.ButtonHeld += MoveRightHeld;
		MoveRight.ButtonReleased += MoveRightReleased;
		MoveDown.ButtonHeld += MoveDownHeld;
		MoveDown.ButtonReleased += MoveDownReleased;

		_audioSource = GetComponent<AudioSource>();
		_audioSource.PlayOneShot(GameStart);

		//Loading of all shapes that was created
		TextAsset[] temp = Resources.LoadAll<TextAsset>("Shapes");
		List<Shape> loadedShapes = new List<Shape>();
		foreach (TextAsset obj in temp)
		{
			string[] textData = obj.text.Split(new string[]{"\r\n"}, StringSplitOptions.RemoveEmptyEntries);
			string name = string.Empty;
			string data = string.Empty;
			for (int i = 0; i < textData.Length; ++i)
			{
				if (i == 0)
					name = textData[i];
				else
				{
					data += textData[i];
				}
			}

			loadedShapes.Add(new Shape(JSONArray.Parse(data), name));
		}
		_gameShapes = loadedShapes.ToArray();

		////Just put a blank clear texture in the next piece for now
		var imageAsset = new Texture2D(1, 1, TextureFormat.ARGB32, false);
		imageAsset.SetPixel(0, 0, Color.clear);
		imageAsset.Apply();
		var spriteAsset = Sprite.Create(imageAsset, new Rect(0, 0, imageAsset.width, imageAsset.height), Vector2.zero, 1);
		NextPiece.sprite = spriteAsset;

		_delayMoveDown = DelayBeforeMoveDown;

		GrabNewShape(_gameShapes[UnityEngine.Random.Range(0, _gameShapes.Length)]);

		StartCoroutine(LoadLogo(LogoUrl));
	}

	/// <summary>
	/// Loading the logo from a given url
	/// </summary>
	/// <param name="url">URL of the logo</param>
	private IEnumerator LoadLogo(string url)
	{
		WWW imageUrl = new WWW(url);
		yield return imageUrl;
		if (imageUrl.texture != null)
			Logo.sprite = Sprite.Create(imageUrl.texture, new Rect(0, 0, imageUrl.texture.width, imageUrl.texture.height), new Vector2(0.5f, 0.5f));
		else
            Logo.sprite = Resources.Load<Sprite>("Logo/bf_icon");
	}

	/// <summary>
	/// Pull the highscores
	/// </summary>
	public void PullHighscores()
	{
		StartCoroutine(LoadHighscores(HighscoresURL));
	}

	/// <summary>
	/// Pull the highscores from a given url
	/// </summary>
	/// <param name="url">URL to pull from</param>
	private IEnumerator LoadHighscores(string url)
	{
		WWW highscoreUrl = new WWW(url);
		yield return highscoreUrl;

		if (!string.IsNullOrEmpty(highscoreUrl.text))
		{
			JSONObject temp = JSONObject.Parse(highscoreUrl.text);
			JSONArray playerScores = temp.GetArray("HighScores");
			string finalText = string.Empty;
			foreach (JSONValue v in playerScores)
			{
				JSONObject innerObj = v.Obj;
				finalText += innerObj["name"].Str + " - " + innerObj["score"] + System.Environment.NewLine;
			}
			Highscores.text = finalText;
		}
		ToggleHighscores(true);
	}

	/// <summary>
	/// Highscores obj to toggle
	/// </summary>
	/// <param name="isVisible">Whether it is visible or not</param>
	public void ToggleHighscores(bool isVisible)
	{
		HighscoresGO.SetActive(isVisible);
	}

	/// <summary>
	/// Restarts the tetris game
	/// </summary>
	public void RestartGame()
	{
#if UNITY_EDITOR
		Debug.Log("Restarted game");
#endif
#if UNITY_4_6 || UNITY_4_7
		Application.LoadLevel(0);
#else
        UnityEngine.SceneManagement.SceneManager.LoadScene(0);
#endif
	}

	/// <summary>
	/// Loads the next shape
	/// </summary>
	/// <param name="nextShape">Shape to display in the top left</param>
	public void LoadNext(Shape nextShape)
	{
		_nextShape = nextShape;
		var imageAsset = new Texture2D(1, 1, TextureFormat.ARGB32, false);
		imageAsset.Resize(nextShape.Size, nextShape.Size);

		for (int y = 0; y < imageAsset.height; ++y)
		{
			for (int x = 0; x < imageAsset.width; ++x)
			{
				if (nextShape.ShapeLines[y, x] != 0)
					imageAsset.SetPixel(x, y, Color.white);
				else
					imageAsset.SetPixel(x, y, Color.clear);
			}
		}

		imageAsset.Apply(); //Save the information stored
		TextureScale.Point(imageAsset, 64, 64); //Rescales it to point
		var spriteAsset = Sprite.Create(imageAsset, new Rect(0, 0, imageAsset.width, imageAsset.height), Vector2.zero, 64);
		NextPiece.sprite = spriteAsset;
	}

	/// <summary>
	/// Rotate the current shape
	/// </summary>
	public void RotateShape()
	{
		//Get the new rotation matrix
		int[,] newMatrix = RotateMatrix(_currentShape.ShapeLines, _currentShape.Size);
		
		//Found the location of this shape!
		//If we can rotate, then do so, otherwise ignore!
		bool cleared = true;
		int offsetX = 0;
		int offsetY = 0;
		for (int y = _shapeOffsetY; y < GameTetrisLines.Count; ++y)
		{
			if (offsetY >= _currentShape.Size)
				break;

			offsetX = 0;
			for (int x = _shapeOffsetX; x < 7; ++x)
			{
				if (offsetX >= _currentShape.Size)
					break;

				if (newMatrix[offsetY, offsetX] != 0)
				{
					if (!GameTetrisLines[y].IsBlockEmpty(x, _prevShapeID))
					{
						cleared = false;
					}
				}
				offsetX++;
			}
			offsetY++;
		}

		if (!cleared)
		{
#if UNITY_EDITOR
			Debug.Log("Can't rotate!");
#endif
			return;
		}

		foreach (ShapeObject so in shapeParts)
		{
			so.ClearShape();
		}
		shapeParts.Clear();

		_currentShape.ShapeLines = newMatrix;
		offsetY = 0;
		for (int y = _shapeOffsetY; y < GameTetrisLines.Count; ++y)
		{
			if (offsetY >= _currentShape.Size)
				break;

			offsetX = 0;
			for (int x = _shapeOffsetX; x < 7; ++x)
			{
				if (offsetX >= _currentShape.Size)
					break;

				if (newMatrix[offsetY, offsetX] != 0)
				{
					shapeParts.Add(new ShapeObject(GameTetrisLines[y], x, y, _currentShape.Size, _prevShapeID, _prevShapeColor));
				}
				offsetX++;
			}
			offsetY++;
		}
	}

	public int[,] RotateMatrix(int[,] matrix, int size)
	{
		int[,] ret = new int[size, size];

		for (int i = 0; i < size; ++i)
		{
			for (int j = 0; j < size; ++j)
			{
				ret[i, j] = matrix[size - j - 1, i];
			}
		}

		return ret;
	}

	/// <summary>
	/// Move the shape to the left
	/// </summary>
	public void MoveLeftHeld()
	{
		_isMovingLeft = true;
		_moveDelayTimer = 0.45f;

#if UNITY_EDITOR
		Debug.Log("Moving left");
#endif

		//Move left
		MoveBlocksLeft();
	}

	/// <summary>
	/// Stop moving shape to the left
	/// </summary>
	public void MoveLeftReleased()
	{
		_isMovingLeft = false;
	}

	/// <summary>
	/// Move the shape to the right
	/// </summary>
	public void MoveRightHeld()
	{
		_isMovingRight = true;
		_moveDelayTimer = 0.45f;

#if UNITY_EDITOR
		Debug.Log("Moving right");
#endif

		//Move right
		MoveBlocksRight();
	}

	/// <summary>
	/// Stop moving the shape to the right
	/// </summary>
	public void MoveRightReleased()
	{
		_isMovingRight = false;
	}

	/// <summary>
	/// Move the shape downwards
	/// </summary>
	public void MoveDownHeld()
	{
		DelaySpeed = 5;
	}

	/// <summary>
	/// Stop moving the shape downwards
	/// </summary>
	public void MoveDownReleased()
	{
		DelaySpeed = 1;
	}

	/// <summary>
	/// Update for all input in the game
	/// </summary>
	public void Update()
	{
		if (_gameOver)
			return;

		if (Input.GetKeyDown(KeyCode.RightArrow))
		{
			MoveRightHeld();
		}
		else if (Input.GetKeyUp(KeyCode.RightArrow))
		{
			MoveRightReleased();
		}

		if (Input.GetKeyDown(KeyCode.LeftArrow))
		{
			MoveLeftHeld();
		}
		else if (Input.GetKeyUp(KeyCode.LeftArrow))
		{
			MoveLeftReleased();
		}

		if (Input.GetKeyDown(KeyCode.DownArrow))
		{
			MoveDownHeld();
		}
		else if (Input.GetKeyUp(KeyCode.DownArrow))
		{
			MoveDownReleased();
		}

		if (Input.GetKeyDown(KeyCode.UpArrow))
		{
			RotateShape();
		}

		if (_isMovingLeft && !_isMovingRight)
		{
			//Move current shape to the left
			if (_moveDelayTimer > 0)
			{
				_moveDelayTimer -= Time.deltaTime;
			}
			else
			{
				_moveDelayTimer = 0.45f;
				//Move left
				MoveBlocksLeft();
			}
		}
		else if (_isMovingRight && !_isMovingLeft)
		{
			//Move current shape to the right
			if (_moveDelayTimer > 0)
			{
				_moveDelayTimer -= Time.deltaTime;
			}
			else
			{
				_moveDelayTimer = 0.45f;
				//Move right
				MoveBlocksRight();
			}
		}

		if (_delayMoveDown > 0)
		{
			_delayMoveDown -= Time.deltaTime * DelaySpeed;
		}
		else
		{
			_delayMoveDown = DelayBeforeMoveDown;

			//Move the block down if possible
			bool canMoveDown = true;

			//Optimize Later
			/// Organize the shape to be properly positions with the lowers y being called first,
			/// and the highest being called last
			List<ShapeObject> orderedList = shapeParts.OrderBy(o => o.PositionY).ToList();

			for (int i = 0; i < orderedList.Count; ++i)
			{
				if (orderedList[i].NextHeight < 0)
					canMoveDown = false;
				else if (!orderedList[i].MoveDown(GameTetrisLines[orderedList[i].NextHeight]))
					canMoveDown = false;
			}
			if (canMoveDown)
			{
				_shapeOffsetY -= 1;

				for (int i = 0; i < orderedList.Count; ++i)
				{
					orderedList[i].FinalizeMove(GameTetrisLines[orderedList[i].PositionY - 1]);
				}
			}
			else
			{
				///Determine if a line is full, if so, clear and add points
				for (int i = GameTetrisLines.Count - 1; i >= 0; --i)
				{
					if (GameTetrisLines[i].isLineFilled())
					{
						_score += 1000;
						Score.text = _score.ToString("0000000");
						GameTetrisLines[i].ClearLine();
						_audioSource.PlayOneShot(ClearedLine);

						//Iterate through all above lines and move them down.
						for (int j = i; j < GameTetrisLines.Count; ++j)
						{
							int aboveLine = j + 1;

							if (aboveLine >= GameTetrisLines.Count)
								break;

							GameTetrisLines[j].CloneLine(GameTetrisLines[aboveLine]);

							if (j + 1 >= GameTetrisLines.Count - 1)
								GameTetrisLines[j + 1].ClearLine();
						}
					}
				}

#if UNITY_EDITOR
				Debug.Log("Grabbing New Shape!");
#endif
				GrabNewShape(_nextShape);
			}
		}
	}

	/// <summary>
	/// Gets a random new shape, stops input for previous parts
	/// </summary>
	/// <param name="nextShape">What shape to use next</param>
	public void GrabNewShape(Shape nextShape)
	{
		_currentShape = nextShape;
		LoadNext(_gameShapes[UnityEngine.Random.Range(0, _gameShapes.Length)]);
		shapeParts.Clear();

		_shapeOffsetX = 2;
		_shapeOffsetY = GameTetrisLines.Count - _currentShape.Size;

		bool Clear = true;
		if (GameTetrisLines[GameTetrisLines.Count - 1].IsBlockEmpty(3, DEFAULT_SHAPE_ID))
		{
			for (int i = GameTetrisLines.Count - 1; i >= GameTetrisLines.Count - _currentShape.Size; --i)
			{
				int dimension = GameTetrisLines.Count - (i + 1);

				for (int x = 0; x < _currentShape.Size; ++x)
				{
					if (_currentShape.ShapeLines[dimension, x] == 0)
						continue;

					if (!GameTetrisLines[i].IsBlockEmpty(x + OffsetX, DEFAULT_SHAPE_ID))
					{
						//Line unavailable for this shape
#if UNITY_EDITOR
						Debug.LogError("Blocked!");
#endif
						Clear = false;
					}
				}
			}
		}
		else
		{
			Clear = false;
		}

		if (Clear)
		{
			if (GameTetrisLines[GameTetrisLines.Count - 1].IsBlockEmpty(3, DEFAULT_SHAPE_ID))
			{
				Color randColor = GrabRandomColor(_shapeID);

				_prevShapeColor = randColor;
				_prevShapeID = _shapeID;

				for (int i = GameTetrisLines.Count - 1; i >= GameTetrisLines.Count - _currentShape.Size; --i)
				{
					int dimension = GameTetrisLines.Count - (i + 1);
					for (int x = 0; x < _currentShape.Size; ++x)
					{
						if (_currentShape.ShapeLines[dimension, x] == 0)
							continue;

						shapeParts.Add(new ShapeObject(GameTetrisLines[i], x + OffsetX, i, _currentShape.Size, _shapeID, randColor));
					}
				}
				_shapeID++;
			}

			_shapeOffsetY = GameTetrisLines.Count - _currentShape.Size;
		}
		else
		{
			_audioSource.PlayOneShot(GameOver);
#if UNITY_EDITOR
			Debug.Log("Game Over");
#endif
			_gameOver = true;
			GameOverGO.SetActive(true);
		}
	}

	/// <summary>
	/// Move the blocks to the left
	/// </summary>
	public void MoveBlocksLeft()
	{
		if (shapeParts.Count <= 0)
			return;

		bool canMoveLeft = true;

		List<ShapeObject> orderedList = shapeParts.OrderBy(o => o.PositionX).ToList();

		for (int i = 0; i < orderedList.Count; ++i)
		{
			if (!orderedList[i].MoveLeft())
			{
				canMoveLeft = false;
			}
		}

		if (canMoveLeft)
		{
			_shapeOffsetX -= 1;
			for (int i = 0; i < orderedList.Count; ++i)
			{
				orderedList[i].FinalizeMoveLeft();
			}
		}
	}

	/// <summary>
	/// Move the blocks to the right
	/// </summary>
	public void MoveBlocksRight()
	{
		if (shapeParts.Count <= 0)
			return;

		bool canMoveRight = true;

		List<ShapeObject> orderedList = shapeParts.OrderBy(o => o.PositionX).ToList();

		for (int i = 0; i < orderedList.Count; ++i)
		{
			if (!orderedList[i].MoveRight())
			{
				canMoveRight = false;
			}
		}

		if (canMoveRight)
		{
			_shapeOffsetX += 1;

			for (int i = orderedList.Count - 1; i >= 0; --i)
			{
				orderedList[i].FinalizeMoveRight();
			}
		}
	}

	/// <summary>
	/// Gets a random color based off a seed passed in
	/// </summary>
	/// <param name="seed">Seed to be used</param>
	/// <returns></returns>
	public Color GrabRandomColor(int seed)
	{
		UnityEngine.Random.seed = seed;

		Color finalColor = Color.white;
		finalColor.r = UnityEngine.Random.value;
		finalColor.b = UnityEngine.Random.value;
		finalColor.g = UnityEngine.Random.value;

		return finalColor;
	}
}

///////////////////////////////////////
//
//  Copyright 2014-2016 Brett Faulds
//
///////////////////////////////////////